/////////////////////////////////////////////////////////////
// Schmoo2Spreadsheet - Generate a text schmoo file
//
// Authors         : Ken Stevens <kstevens@ece.utah.edu>
// Mod Date        : 5 April 2008
// Revision        : 0.1
// Status          : 
// Bug Reports     : Ken Stevens <kstevens@ece.utah.edu>
//
// This class will parse an LVS500 Schmoo binary data file
// and turn it into a text file suitable for a document or
// a spreadsheet for formatting.
//
// The format was reverse engineered without documentation
// from sample schmoo files, so there still may be some bugs.
// 
// Bugs:
//   Jun Bok You extended the parsing to include more binary
//   characters that are possible between descriptors.
//
// Improvements:
//
// Comments:
//
// Commands:
//
///////////////////////////////////////////////////////////////



#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>



#include "schmoo2spreadsheet.h"



int main(int argc, char* argv[])
{
  FILE *fp;
  unsigned int idx, idx2;
  int c, val0, val1;
  char buf[128]; //, buf1[32], buf2[32];
  //char *cp;
  Schmoo *schmoo = new Schmoo;

  // test for two args: program name and the file to parse
  if (argc < 2 || argc > 3) {
    printf("\nUsage: %s <FILENAME> TAB_DELIM\n\nwhere <FILENAME> is the name of the SCMOO\nplot file from the LV500 tester\n\nTAB_DELIM is an optional argument that\nwhen given, makes tab delimited format suitable for import into excel\n\n", *argv);
    return 1;
  }

  // if we can't open the file, then exit.
  if ((fp = fopen(argv[1], "r")) == NULL) {
    printf("\nError opening file: `%s'\n\n", argv[1]);
    return 1;
  }

  // Read file:  starts with a SOH character (0x01) and the word Schmoo
  if ((c = fscanf(fp, "%[Schmo]", buf)) != 1 || strcmp(buf, "Schmoo")) {
    printf("\nError %d: `%s' doesn't appear to be a schmoo file\n\n", c, argv[1]);
    goto done;
  }

  // Skip to next field we care about: the date
  do c = getc(fp);
  while (c != EOF && !isalpha(c));
  ungetc(c, fp);
  // Read date
  c = fscanf(fp, "%[a-zA-Z] %d, %4d", buf, &val0, &val1);
  if (c != 3) {
    printf("\nERROR: only %d of 3 args found for schmoo date\n", c);
    goto parse_err;
  }
  schmoo->setDate(buf, val0, val1);
  // Skip to time field and read time
  do c = getc(fp);
  while (c != EOF && c != '\0');
  c = fscanf(fp, "%[0-9: APM]", buf);
  if (c != 1) {
    printf("\nERROR: AM/PM of schmoo date not found.\n");
    goto parse_err;
  }
  schmoo->setTime(buf);

  // Get x and y descriptor fields
  for (idx = 0; idx < 2; idx++) {

    // skip to descriptor field
    do c = getc(fp);
    while (c != EOF && !isalpha(c));
    ungetc(c, fp);
    c = fscanf(fp, "%[-A-Za-z0-9. /_]", buf);
    if (c != 1) {
      printf("\nERROR: Schmoo %c descriptor field not found\n", idx ? 'Y' : 'X');
      goto parse_err;
    }
    if (idx)
      schmoo->y->setTitle(buf);
    else
      schmoo->x->setTitle(buf);

    // Get all data values
    while (1) {
      do c = getc(fp);
      // Looks like I have ^@ and ^F
      while (c != EOF && (!c || c == 6 || ((c <=  26) && idx == 0))) ;
      ungetc(c, fp);
      if (isalpha(c) || (idx && (c == PASSED || c == FAILED || c == BOGUS)))
	break;
      c = fscanf(fp, "%[A-Za-z0-9. /_]", buf);
      if (c != 1) {
	printf("\nERROR: Schmoo %c range descriptor field #%d not found\n",
	       idx ? 'Y' : 'X', idx ? schmoo->y->size() : schmoo->x->size());
	if (!idx) {
	  printf("Possible control character that isn't ^@ or ^(any alphabet)??\n"); 
	  printf("Look at between X range descriptor and Y range descriptor\n");
	} else {
	  printf("Possible control character that isn't ^@ or ^F??\n");
	  printf("Look at after Y range descriptor\n");
	}
	goto parse_err;
      }
      if (idx)
	schmoo->y->addValue(buf);
      else
	schmoo->x->addValue(buf);
    }
  }


  // Init schmoo data structure
  schmoo->initData();

  // Get data row for each y value
  for (idx = 0; idx < schmoo->x->size(); idx++) {
    for (idx2 = 0; idx2 < schmoo->y->size(); idx2++) {
      c = getc(fp);
      if (c != PASSED && c != FAILED && c != BOGUS) {
	printf("\nERROR: bogus schmoo value: %d location x,y = %d,%d\n",
	       c, idx, idx2);
	goto parse_err;
      }
      schmoo->setData(idx, idx2, (SCHMOOTYPE) c);
    }
    do c = getc(fp);
    while (c != EOF && !c);
    ungetc(c, fp);
  }


  printf("\nSchmoo Plot: %s %d, %d %s\n\n", schmoo->month(), schmoo->day(),
	 schmoo->year(), schmoo->time());

//    printf("X Title: %s\n  size: %d\n", schmoo->x->title(), schmoo->x->size());
//    for (idx = 0; idx < schmoo->x->size(); idx++)
//      printf("%s  ", schmoo->x->value(idx));
//  
//    printf("\nY Title: %s\n  size: %d\n", schmoo->y->title(), schmoo->y->size());
//    for (idx = 0; idx < schmoo->y->size(); idx++)
//      printf("%s  ", schmoo->y->value(idx));
//    printf("\n\n");

  schmoo->print(argc == 3);

  goto done;

 parse_err:
  printf("\nError parsing file\n\n");
  return 0;

 done:
  fclose(fp);
  return 1;

}







///////////////////////////////////////////////////////////////////////////
//   Schmoo data structure functions
///////////////////////////////////////////////////////////////////////////



//-------------------------------------------
// Schmoo::Schmoo
// Schmoo constructor. Initializes member variables
// Inputs: None
// Outputs: None
Schmoo::Schmoo()
{
  month_ = time_ = NULL;
  day_ = year_ = 0;
  x = new SchmooData;
  y = new SchmooData;
  data_ = (SCHMOOTYPE **)NULL;
}


//-------------------------------------------
// Schmoo::~Schmoo
// Schmoo destructor. Free malloc'ed members
// Inputs: None
// Outputs: None
Schmoo::~Schmoo()
{
  if (month_)
    free(month_);
  if (time_)
    free(time_);

  free(data_);

  // delete x,y data
  delete(x);
  delete(y);
}



void Schmoo::print(int tabp)
{
  unsigned int idx, idx2;
  SCHMOOTYPE v;

  // print y index values and array data
  for (idx = 0; idx < y->size(); idx++) {
    for (idx2 = 0; idx2 < x->size(); idx2++) {
      // We read from low voltage to high, but print from high voltage to low
      if (!idx2)
	printf("%s", y->value(y->size() - (idx + 1)));
      if (!idx2 || tabp)
	putchar('\t');
      v = getData(idx2, y->size() - (idx + 1));
      printf("%c", v == PASSED ? PASSEDC : v == FAILED ? FAILEDC : BOGUSC);
    }
    printf("\n");
  }

  // Print x index values
  if (tabp)
    for (idx2 = 0; idx2 < x->size(); idx2++)
      printf("\t%s", x->value(idx2));
  else
    printf("\t  %s ... %s\n\n", x->value(0), x->value(x->size() - 1));


  printf("X Axis: %s\n\n", x->title());

  printf("Y Axis: %s\n\n", y->title());

  printf("\n\nLegend: %c = bogus, %c = passed, %c = failed\n\n",
	 BOGUSC, PASSEDC, FAILEDC);
}



void Schmoo::setDate(const char *m, int d, int y)
{

  month_ = (char *)malloc(1+ strlen(m));
  strcpy(month_, m);
  day_ = d;
  year_ = y;
}

void Schmoo::setTime(const char *t)
{
  time_ = (char *)malloc(1+ strlen(t));
  strcpy(time_, t);
}


void Schmoo::initData()
{
  unsigned int idx, idx2;

  if (data_) {
    for (idx = 0; idx < y->size(); idx++)
      free(data_[idx]);
    free(data_);
  }

  data_ = (SCHMOOTYPE **)malloc(y->size() * sizeof(SCHMOOTYPE *));
  for (idx = 0; idx < y->size(); idx++) {
    data_[idx] = (SCHMOOTYPE *)malloc(x->size() * sizeof(SCHMOOTYPE));
    for (idx2 = 0; idx2 < x->size(); idx2++)
      data_[idx][idx2] = BOGUS;
  }

}
 



//-------------------------------------------
// SchmooData::~SchmooData
// SchmooData destructor.  Free malloc'ed data
// Inputs: None
// Outputs: None
SchmooData::~SchmooData()
{
  // free title
  if (title_)
    free(title_);

  std::vector<char *>::iterator iter;

  // Free all the malloc'ed strings in the values_ vector
  for (iter = values_.begin(); iter != values_.end(); iter++) {
    free(*iter);
  }
}


